# -*- coding: utf-8 -*-
# Copyright (C) 2020-2021  Nexedi SA and Contributors.
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

# verify general functionality

import sys
import re
import time
from os.path import dirname
from golang import chan, select, default, func, defer
from golang import context, sync

import pytest

from nxdtest import main


@pytest.fixture
def run_nxdtest(tmpdir):
    """Fixture which returns a function which invokes nxdtest in a temporary
    directory, with the provided .nxdtest file content and with arguments
    passed as `argv`.
    """

    def _run_nxdtest(nxdtest_file_content, argv=("nxdtest",)):
        with tmpdir.as_cwd():
            with open(".nxdtest", "w") as f:
                f.write(nxdtest_file_content)
            sys_argv = sys.argv
            sys.argv = argv
            try:
                main()
            finally:
                sys.argv = sys_argv

    return _run_nxdtest


def test_main(run_nxdtest, capsys):
    run_nxdtest(
        """\
TestCase('TESTNAME', ['echo', 'TEST OUPUT'])
"""
    )

    captured = capsys.readouterr()
    output_lines = captured.out.splitlines()
    assert ">>> TESTNAME" in output_lines
    assert "$ echo TEST OUPUT" in output_lines
    assert "TEST OUPUT" in output_lines
    assert re.match("ok\tTESTNAME\t.*s\t# 1t 0e 0f 0s", output_lines[-2])
    assert re.match(u"# ran 1 test case:  1·ok", output_lines[-1])


def test_error_invoking_command(run_nxdtest, capsys):
    run_nxdtest(
        """\
TestCase('TESTNAME', ['not exist command'])
"""
    )

    captured = capsys.readouterr()
    assert "No such file or directory" in captured.err


def test_error_invoking_summary(run_nxdtest, capsys):
    run_nxdtest(
        """\
TestCase('TESTNAME', ['echo'], summaryf="error")
"""
    )

    captured = capsys.readouterr()
    assert "TypeError" in captured.err


def test_error_execing_nxdtest_file(run_nxdtest, capsys):
    with pytest.raises(ZeroDivisionError) as excinfo:
        run_nxdtest(
        """\
1 / 0
"""
    )
    assert '1 / 0' in str(excinfo.traceback[-1])
    # The actual .nxdtest filename is also included in traceback
    assert ".nxdtest':1" in str(excinfo.traceback[-1])


def test_run_argument(run_nxdtest, capsys):
    run_nxdtest(
        """\
TestCase('TEST1', ['echo', 'TEST1'])
TestCase('TEST2', ['echo', 'TEST2'])
""",
        argv=["nxdtest", "--run", "TEST1"],
    )
    captured = capsys.readouterr()
    assert "TEST1" in captured.out
    assert "TEST2" not in captured.out

    # --run uses search
    run_nxdtest(
        """\
TestCase('TEST1', ['echo', 'TEST1'])
TestCase('TEST2', ['echo', 'TEST2'])
TestCase('TEST10', ['echo', 'TEST10'])
""",
        argv=["nxdtest", "--run", "ST1"],
    )
    captured = capsys.readouterr()
    assert "TEST1" in captured.out
    assert "TEST10" in captured.out
    assert "TEST2" not in captured.out


# verify that nxdtest detects leaked processes.
@pytest.mark.timeout(timeout=10)
def test_run_procleak(run_nxdtest, capsys):
    procleak = "%s/testprog/procleak" % (dirname(__file__),)

    # run nxdtest in thread so that timeout handling works
    # ( if nxdtest is run on main thread, then non-py wait in WorkGroup.wait, if
    #   stuck, prevents signals from being handled at python-level )
    wg = sync.WorkGroup(context.background())
    done = chan()
    @func
    def _(ctx):
        defer(done.close)

        run_nxdtest(
            """\
TestCase('TEST_WITH_PROCLEAK', ['%s', 'AAA', 'BBB', 'CCC'])
"""     % procleak
        )
    wg.go(_)

    while 1:
        _, _rx = select(
                    default,    # 0
                    done.recv,  # 1
        )
        if _ == 0:
            time.sleep(0.1)
            continue
        wg.wait()
        break

    captured = capsys.readouterr()
    assert "AAA: terminating" in captured.out
    assert "BBB: terminating" in captured.out
    assert "CCC: terminating" in captured.out
